from abc import abstractmethod
from typing import List, Optional, Any, Self

from flask import g
from portal import db


class Retrievable(db.Model):
    """
    A type whose instances can be retrieved from the database.

    It has methods to get all instances and a specific instance by its primary key.
    """

    # See https://flask-sqlalchemy.palletsprojects.com/en/3.0.x/customizing/#abstract-models-and-mixins
    __abstract__ = True

    @classmethod
    def get_canonical_order_column(cls) -> Optional[db.Column]:
        """
        Return the colum by which instances of this type should be canonically ordered
        when retrieving them all from the database.

        By default, the column is None, i.e. no order is enforced.
        """
        return None

    @classmethod
    def get_all(cls) -> List[Self]:
        """
        Return all instances of this type.

        Cached per request: After the first call of get_all(), all instances are stored in the request globals
        object, i.e. not retrieved again _for that particular request_.
        """
        if "instances" not in g or not g.instances.get(cls):
            g.setdefault("instances", {})[cls] = {
                instance.primary_key: instance
                for instance in db.session.execute(
                    db.select(cls).order_by(cls.get_canonical_order_column())
                ).scalars()
            }

        return list(g.instances[cls].values())

    @classmethod
    def get_only(cls, primary_key: Any) -> Optional[Self]:
        """
        Return the instance of this type with the given primary key or None if it doesn't exist.

        If the cache has been populated using get_all(), the instance is retrieved from cache.
        """
        if "instances" in g and g.instances.get(cls):
            return g.instances[cls].get(primary_key)
        else:
            return db.session.get(cls, primary_key)

    @property
    def primary_key(self) -> Any:
        """
        Return the primary key that can be used to identify an instance of this class.
        """
        raise NotImplementedError(
            "The primary key property needs to be implemented by subclasses."
        )

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """
        Return a dictionary representation of this instance.

        Note that this representation might not be complete (and instead only contain fields that are currently
        relevant for at least one purpose), so do not rely on it if you want a complete serialization of this
        object.
        """
        pass
